/*->c.Drawlevel0 */

/* -> c.DrawLevel0
 *
 * DrawFile module, level 0 (diagram level interface)
 * History:
 * Version 0.0: 21 Nov 88, DAHE: created
 *         0.1: 23 Nov 88, DAHE: first working version
 *         0.2: 24 Nov 88, DAHE: added shift and query
 *         0.3: 28 Nov 88, DAHE: name changes, box conversion
 *         0.4: 16 Feb 89, DAHE: unknown object handler added
 *
 * Implements the interface as seen by the outside world to the draw file
 * module, at the level of diagrams.
 *
 * How to use the module (the simplest possible use)
 *  allocate a block of memory and set a Diag to point to it.
 *  read a Draw file into the block and set the length.
 *  (maybe) call Draw_verify_diag.
 *  call Draw_render_diag.
 */


#include "stdafx.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "os.h"
#include "sprite.h"
#include "wimp.h"
#include "font.h"


#include "transform.h"
#include "xprocess.h"

#include "drawmod.h"


#define DRAWFILE_INTERNAL
#include "DrawFile1.h"
#include "DrawLevel0.h"
#include "DrawLevel1.h"
#include "DrawFile2.h"
#include "DrawErrors.h"
#include "Drawrebind.h"




/*------------------------ Version string ---------------------------*/
/* The following literal string simply appears in the object file;
   nothing uses it.
   It contains the release version and date.
   Note that it is the version number of the module of a whole, not of this
   file.
*/
/*

char *Draw_versionString = "Release 0.2, 15 Feb 89";

 */

/*----------------------- Global data --------------------------------*/

/* Memory allocation/freeing functions */
Draw_allocate Draw_allocator = NULL;
Draw_extend   Draw_extender  = NULL;
Draw_free     Draw_freer     = NULL;

/* Unknown object handler */
Draw_unknown_object_handler dr_unknown_handler;

Draw_enhanced_object_handler dr_enhanced_handler;


/*------------------------------ Internal functions ----------------------------*/
/* These functions are local to the module */

/*
 Function    : dr_spaceCopy
 Purpose     : copy memory
 Parameters  : start of area to copy to
               start of area to copy from
               length of block
 Returns     : void
 Description : copies the block of memory - works even if blocks overlap.
*/

void dr_spaceCopy(char *to, char *from, int length)
{
  if (length <= 0) return;
  if (to < from)
    memcpy((void*)to, (void*)from, length);
  else if (to > from)
  { char *f = from + length - 1;
    char *t = f + (int)(to - from);
    char *limit = from;

    while (f >= limit) *t-- = *f--;
  }
}


/*
 Function    : dr_findFontTable
 Purpose     : find the font table object in a diagram
 Parameters  : diagram structure
               draw object pointer: set of font table found
 Returns     : TRUE if found
*/

BOOL dr_findFontTable(Draw_diag * diag, draw_objptr *hdrptr)
{
  char *limit  = diag->data + diag->length;

  for (hdrptr->bytep = diag->data + sizeof(draw_fileheader);
       hdrptr->bytep < limit; hdrptr->bytep += hdrptr->objhdrp->size)
  {
    if (hdrptr->objhdrp->tag == draw_OBJFONTLIST) return (TRUE);
  }

  return (FALSE);
}

/*
 Function    : dr_mergeFontTables
 Purpose     : merge font tables and convert text references
 Parameters  : pointer to first diagram structure
               object header for first font table [may be non-existent]
               second diagram structure
               object header for second font table [must exist]
               number translation table
               OUT: error block
 Returns     : TRUE if successful
 Description : the two font tables are merged; this can change the references to
               fonts in the second diagram, so we record the changes in the
               translation table. The translation table in indexed by the old
               font reference and yields the new one.
               If this yields too many font refences, we report an error
*/

BOOL dr_mergeFontTables(Draw_diag *diag1, draw_objptr fontTable1,
                               Draw_diag diag2,  draw_objptr fontTable2,
                               int *table, Draw_error *error)
{
  int  fontref, maxref = -1;
  char *names[256]; /* Pointer to names to place in final table */
  char *from, *limit, *to;
  int  size, newsize, sizeChange; /* In bytes */

  for (fontref = 0 ; fontref < 256 ; fontref++) names[fontref] = NULL;

  if (fontTable1.bytep)
  {
    /* Copy first table into names array */
    from  = &(fontTable1.fontlistp->fontref);
    limit = fontTable1.bytep + fontTable1.fontlisthdrp->size;
    while (from < limit)
    {
      if ((fontref = (int)(*from++)) == 0) break;
      names[fontref] = from;
      from += strlen(from)+1;
      if (fontref > maxref) maxref = fontref;
    }
  }

  /* Scan second table */
  from  = &(fontTable2.fontlistp->fontref);
  limit = fontTable2.bytep + fontTable2.fontlisthdrp->size;
  while (from < limit)
  { int oldref, newref;
    BOOL found = FALSE;

    if ((fontref = (int)(*from++)) == 0) break;

    /* Look for diag2 font in diag1 table */
    for (oldref = 1 ; oldref <= maxref ; oldref++)
    {
      if (names[fontref] != NULL && strcmp(names[oldref], from) == 0)
      { /* Font in diagram 2 also exists in diagram 1 - not reference number */
        table[fontref] = oldref;
        found = TRUE;
        break;
      }
    }

    if (!found)
    {
      /* Create new entry */
      for (newref = 1 ; newref < 256 ; newref++)
      {
        if (names[newref] == NULL)
        {
          names[newref]  = from;
          table[fontref] = newref;
          break;
        }
      }

      if (newref == 256)
      {
        /* Ran out of space in new table: give up */
        error->type = DrawOwnError;
        error->err.draw.code     = Draw_TooManyFonts;
        error->err.draw.location = (int)(from - diag2.data);
        return (FALSE);
      }
    }

    from += strlen(from)+1;
  }

  /* Count size of new font table */
  for (newsize = sizeof(draw_fontliststrhdr), fontref = 1 ;
       fontref < 256 ; fontref++)
  {
    if (names[fontref]) newsize += strlen(names[fontref]) + 2;
  }
  newsize = (newsize + sizeof(int) - 1) & (-(int)sizeof(int));

  /* See if the size has changed : always >= 0 */
  if (fontTable1.bytep)
  {
    sizeChange = newsize - (size = fontTable1.fontlisthdrp->size);
  }
  else
  {
    sizeChange = newsize;
    fontTable1.bytep = diag1->data + sizeof(draw_fileheader);
    size = 0;
  }

  if (sizeChange > 0)
  { /* Move rest of memory down */
    char *from = fontTable1.bytep + size;
    dr_spaceCopy(fontTable1.bytep + newsize, from,
                 (int)(diag1->data + diag1->length - from));

    /* Build new font table */
    fontTable1.fontlisthdrp->size = newsize;
    to = &(fontTable1.fontlistp->fontref);
    for (fontref = 1 ; fontref < 256 ; fontref++)
    {
      if (names[fontref])
      {
        *to++ = (char)fontref;
        strcpy(to, names[fontref]);
        to += strlen(to);
        *to++ = '\0';
      }
    }

    /* Add final nulls */
    limit = fontTable1.bytep + fontTable1.fontlistp->size;
    while (to < limit) *to++ = '\0';
    fontTable1.objhdrp->tag = draw_OBJFONTLIST; /* In case we created it */

    /* Record new diagram size */
    diag1->length += sizeChange;
  }
  return (TRUE);
}

/*
 Function    : dr_unifyBoxes
 Purpose     : unify two boxes
 Parameters  : pointer to box 1 structure
               pointer to box 2 structure
 Returns     : void
 Description : box1 is set to the union of box 1 and box 2.
*/

void dr_unifyBoxes(draw_bboxtyp *box1, draw_bboxtyp *box2)
{
  if (box2->x0 < box1->x0) box1->x0 = box2->x0;
  if (box2->y0 < box1->y0) box1->y0 = box2->y0;
  if (box2->x1 > box1->x1) box1->x1 = box2->x1;
  if (box2->y1 > box1->y1) box1->y1 = box2->y1;
}

/*-------------------- Interface functions ---------------------------*/


/*
 Function    : Draw_verify_diag
 Purpose     : check a diagram for errors
 Parameters  : diagram structure
               OUT: error block
 Returns     : TRUE on error
 Description : Checks a diagram after it has been read in from a file. Each
               object in the file is verified, and the first error returned.
*/


BOOL Draw_verify_diag(Draw_diag diag,Draw_error *error)
{
 int code;
 int location;


 drawtypemask=
   (!strncmp(((draw_fileheader*)diag.data)->progident,"DrawPlus",8) ||
    !strncmp(((draw_fileheader*)diag.data)->progident,"Vector",6))?0xFF:-1;


 if(check_Draw_file((char *)diag.data,diag.length,&code,&location))
 {
  Draw_rebind_check(&diag);
  return(TRUE);
 }
 else
 {
  DrawFile_Error(code,location);
  return (FALSE);
 }
}



/*
 Function    : Draw_append_diag
 Purpose     : append one diagram to another
 Parameters  : pointer to structure for diagram to append to
               structure for diagram to be appended
               OUT: error block
 Returns     : TRUE if successful
 Description : Merge diag2 into diag1. Both diagrams must have been passed
               through Draw_verify_diag() first. The data block for diag1 must
               be at least diag1.length+diag2.length bytes long. The length
               field of diag1 is updated to the actual merged size which will
               be at most equal to this sum.
               The bounding box is diag1 is set to the union of the bounding
               boxes of the two diagrams.
               Note: after this call, the offsets of objects in diag1 may have
               changed, because the font table object can change in size. No
               other objects will change. Any errors which refer to a specific
               location are in diag2.
*/


#ifdef BIGD

BOOL Draw_append_diag(Draw_diag *diag1, Draw_diag diag2, Draw_error *error)
{
  draw_objptr fontTable1, fontTable2;
  Draw_objectType header1, header2;
  header1.bytep = diag1->data;
  header2.bytep = diag2.data;

  if (dr_findFontTable(diag2, &fontTable2))
  { draw_objptr hdrptr, dest;
    char *end = diag2.data + diag2.length;
    int transTable[256];
    int i, length;

    for (i = 0 ; i < 256 ; i++) transTable[i] = 0;

    if (!dr_findFontTable(*diag1, &fontTable1)) fontTable1.bytep = 0;
    if (!dr_mergeFontTables(diag1, fontTable1, diag2, fontTable2, transTable,
        error))
      return (FALSE);

    /* Copy all objects from diag2 to diag1, except font table(s) */
    dest.bytep  = diag1->data + diag1->length;
    for (hdrptr.bytep = diag2.data + sizeof(draw_fileheader) ;
         hdrptr.bytep < end ; hdrptr.bytep += length)
    {
      length = hdrptr.objhdrp->size;
      if (hdrptr.objhdrp->tag != draw_OBJFONTLIST)
      {
        memcpy(dest.bytep, hdrptr.bytep, length);

        if (dest.objhdrp->tag == draw_OBJTEXT)
        { /* Translate font number */
          dest.textp->textstyle.fontref = transTable[dest.textp->textstyle.fontref];
        }

        dest.bytep += length;
      }
    }
    diag1->length = (int)(dest.bytep - diag1->data);
    dr_unifyBoxes(&header1.fileHeader->bbox, &header2.fileHeader->bbox);
    return (TRUE);
  }
  else
  { /* No font table in diag2 - just copy whole thing */
    int length = diag2.length - sizeof(draw_fileheader);
    memcpy(diag1->data + diag1->length, diag2.data + sizeof(draw_fileheader),
           length);
    diag1->length += length;
    dr_unifyBoxes(&header1.fileHeader->bbox, &header2.fileHeader->bbox);
    return (TRUE);
  }
}

#endif


/*
 Function    : Draw_render_diag
 Purpose     : render a diagram
 Parameters  : diagram structure
               wimp-style redraw structure
               scale factor
               OUT: error block
 Returns     : TRUE if successful
 Description : Render a diagram, which should have been passed through
               Draw_verify_diag() first, at a factor 'scale' of its definition
               size. The object is rendered into the region of the screen
               indicated by the redrawstr.

               Note: the redrawstr is equivalent to a wimp_redrawstr, which may
               be cast to it. The window handle is NOT used. For applications
               not using the wimp, the structure should be set up as follows:
               the graphics box determines the region in screen units which
               will be drawn. Any objects wholly or partly in this region are
               rendered. Clipping must be set by the caller;
               the part of the diagram rendered is determined by mapping the
               top left of the diagram, in Draw coordinate space, onto a point
               (r->box.x0 - r->scx, r->box.y1 - r->scy), in screen coordinate
               space.

               After the call, the VDU5 character size is left at the default
               for the current mode.
*/


#ifdef BIGD


BOOL Draw_render_diag(Draw_diag * diag,Draw_redrawstr *r,transformstr * tr,                                              processstr * pr,Draw_error *error)
{
 draw_objptr dummy;

 /* Render the objects via the level 1 interface */
 dummy.bytep = 0;
 note_fontcat(dummy);
 return(Draw_doObjects(diag, Draw_FirstObject, Draw_LastObject,r,tr,pr,error));
}

#endif


BOOL Draw_render(Draw_diag * diag,Draw_box * box,transformstr * tr,
                                         processstr * pr,Draw_error *error)
{
 draw_objptr dummy;

 /* Render the objects via the level 1 interface */
 dummy.bytep = 0;
 note_fontcat(dummy);
 return(Draw_renderObjects(diag,Draw_FirstObject,Draw_LastObject,box,
                                                                tr,pr,error));
}


/*
 Function    : Draw_registerMemoryFunctions
 Purpose     : register memory aloocate/free functions
 Parameters  : allocate function
               extend function
               free function
 Returns     : void
 Description : This call is only needed if text area objects are to be
               rendered. It specifies three functions, used to allocate, extend
               and free blocks of memory. The memory is used only during the
               rendering of a text area, and is freed before the rendering code
               returns to the caller.

               If this routine is never called, or if memory allocation fails,
               then attempting to render a text area will produce no output.
*/

void Draw_registerMemoryFunctions(Draw_allocate alloc,
                                  Draw_extend   extend,
                                  Draw_free     free)
{
  Draw_allocator = alloc;
  Draw_extender  = extend;
  Draw_freer     = free;
}

/*
 Function    : Draw_shift_diag
 Purpose     : Shift a diagram
 Parameters  : diagram structure
               shift to apply
 Returns     : void
 Description : All coordinates in the diagram are moved by the given distance.
*/

#ifdef BIGD

void Draw_shift_diag(Draw_diag diag, int xMove, int yMove)
{
  shift_Draw_file(diag.data, diag.length, xMove, yMove);
}

#endif


/*
 Function    : Draw_set_unknown_object_handler
 Purpose     : register unknown object handler
 Parameters  : object handler
 Returns     : previous handler
 Description : Updates all font reference numbers for text objects following
               creation of a font table. See Draw_createObject.
*/

#ifdef BIGD

Draw_unknown_object_handler Draw_set_unknown_object_handler
                           (Draw_unknown_object_handler handler)
{
  Draw_unknown_object_handler oldhandler = dr_unknown_handler;

  dr_unknown_handler = handler;
  return (oldhandler);
}

#endif


Draw_enhanced_object_handler Draw_set_enhanced_object_handler
                           (Draw_enhanced_object_handler handler)
{
  Draw_enhanced_object_handler oldhandler = dr_enhanced_handler;

  dr_enhanced_handler=handler;
  return (oldhandler);
}




os_error * Draw_CacheFonts(Draw_diag * diag,DrawCachefn cache,DrawCacheCharfn cachechar)
{
 os_error * err;
 int        i;
	void     * handle;

 err=NULL;

 Draw_setFontTable(diag);

 if(fontcatseen)
 {
  for(i=0;i<=fontcat.list_size;i++)
  {
   if(fontcat.name[i])
   {
    err=cache(fontcat.name[i],0,&handle);
				if(!err)
				{
					if(cachechar) err=cachechar(32,255,handle);
				}
			}
   if(err) break;
  }
 }

 return(err);
}



Draw_findfn           Draw_find=font_find;
Draw_losefn           Draw_lose=font_lose;
Draw_paintfn          Draw_paint=font_paint;
Draw_painttransfn     Draw_painttrans=font_painttrans;
Draw_setfontfn        Draw_setfont=font_setfont;
Draw_strwidthfn       Draw_strwidth=font_strwidth;
Draw_stringbboxfn     Draw_stringbbox=font_stringbbox;
Draw_setcolourfn      Draw_setcolour=font_setcolour;




void Draw_registerFontFunctions(Draw_findfn find,
                                Draw_losefn lose,
                                Draw_paintfn paint,
                                Draw_painttransfn painttrans,
                                Draw_setfontfn setfont,
                                Draw_strwidthfn strwidth,
                                Draw_stringbboxfn stringbbox,
                                Draw_setcolourfn setcolour)


{
 Draw_find=find;
 Draw_lose=lose;
 Draw_paint=paint;
 Draw_painttrans=painttrans;
 Draw_setfont=setfont;
 Draw_strwidth=strwidth;
 Draw_stringbbox=stringbbox;
 Draw_setcolour=setcolour;
}


extern os_error * rsetgcol(int pal,int action);

Draw_fillfn           Draw_fill=Draw_Fill;
Draw_strokefn         Draw_stroke=Draw_Stroke;
Draw_drawsetcolourfn  Draw_drawsetcolour=rsetgcol;


void Draw_registerDrawFunctions(Draw_fillfn      fill,
                                Draw_strokefn    stroke,
                                Draw_drawsetcolourfn setcolour)
{
 Draw_fill=fill;
 Draw_stroke=stroke;
 Draw_drawsetcolour=setcolour;
}


